package PVE::QemuServer;

use strict;
use POSIX;
use IO::Handle;
use IO::Select;
use IO::File;
use IO::Dir;
use IO::Socket::UNIX;
use File::Basename;
use File::Path;
use File::stat;
use Getopt::Long;
use Digest::SHA1;
use Fcntl ':flock';
use Cwd 'abs_path';
use IPC::Open3;
use Fcntl;
use Sys::Syslog;

#no warnings 'redefine';

unless(defined(&_VZSYSCALLS_H_)) {
    eval 'sub _VZSYSCALLS_H_ () {1;}' unless defined(&_VZSYSCALLS_H_);
    require 'sys/syscall.ph';
    if(defined(&__x86_64__)) {
	eval 'sub __NR_fairsched_vcpus () {499;}' unless defined(&__NR_fairsched_vcpus);
	eval 'sub __NR_fairsched_mknod () {504;}' unless defined(&__NR_fairsched_mknod);
	eval 'sub __NR_fairsched_rmnod () {505;}' unless defined(&__NR_fairsched_rmnod);
	eval 'sub __NR_fairsched_chwt () {506;}' unless defined(&__NR_fairsched_chwt);
	eval 'sub __NR_fairsched_mvpr () {507;}' unless defined(&__NR_fairsched_mvpr);
	eval 'sub __NR_fairsched_rate () {508;}' unless defined(&__NR_fairsched_rate);
	eval 'sub __NR_setluid () {501;}' unless defined(&__NR_setluid);
	eval 'sub __NR_setublimit () {502;}' unless defined(&__NR_setublimit);
    }
    elsif(defined( &__i386__) ) {
	eval 'sub __NR_fairsched_mknod () {500;}' unless defined(&__NR_fairsched_mknod);
	eval 'sub __NR_fairsched_rmnod () {501;}' unless defined(&__NR_fairsched_rmnod);
	eval 'sub __NR_fairsched_chwt () {502;}' unless defined(&__NR_fairsched_chwt);
	eval 'sub __NR_fairsched_mvpr () {503;}' unless defined(&__NR_fairsched_mvpr);
	eval 'sub __NR_fairsched_rate () {504;}' unless defined(&__NR_fairsched_rate);
	eval 'sub __NR_fairsched_vcpus () {505;}' unless defined(&__NR_fairsched_vcpus);
	eval 'sub __NR_setluid () {511;}' unless defined(&__NR_setluid);
	eval 'sub __NR_setublimit () {512;}' unless defined(&__NR_setublimit);
    } else {
	die("no fairsched syscall for this arch");
    }
    require 'asm/ioctl.ph';
    eval 'sub KVM_GET_API_VERSION () { &_IO(0xAE, 0x);}' unless defined(&KVM_GET_API_VERSION);
}

sub fairsched_mknod {
    my ($parent, $weight, $desired) = @_;

    return syscall(&__NR_fairsched_mknod, int ($parent), int ($weight), int ($desired));
}

sub fairsched_rmnod {
    my ($id) = @_;

    return syscall(&__NR_fairsched_rmnod, int ($id));
}

sub fairsched_mvpr {
    my ($pid, $newid) = @_;

    return syscall(&__NR_fairsched_mvpr, int ($pid), int ($newid));
}

sub fairsched_vcpus {
    my ($id, $vcpus) = @_;

    return syscall(&__NR_fairsched_vcpus, int ($id), int ($vcpus));
}

sub fairsched_rate {
    my ($id, $op, $rate) = @_;

    return syscall(&__NR_fairsched_rate, int ($id), int ($op), int ($rate));
}

use constant FAIRSCHED_SET_RATE  => 0;
use constant FAIRSCHED_DROP_RATE => 1;
use constant FAIRSCHED_GET_RATE  => 2;

sub fairsched_cpulimit {
    my ($id, $limit) = @_;

    my $cpulim1024 = int ($limit * 1024 / 100);
    my $op = $cpulim1024 ? FAIRSCHED_SET_RATE : FAIRSCHED_DROP_RATE;

    return fairsched_rate ($id, $op, $cpulim1024);
}

my $confdir = '/etc/qemu-server';
my $imagedir = '/var/lib/vz/images';
my $isodir = '/var/lib/vz/template/iso';


my $var_run_tmpdir = "/var/run/qemu-server";

my $default_memory = 512;

my $confvars = {
    onboot => 'bool',
    autostart => 'bool',
    reboot => 'bool',
    cpulimit => 'natural',
    cpuunits => 'natural',
    memory => 'natural',
    keyboard => 'lang',
    name => 'string',
    description => 'string',
    ostype => 'ostype',
    boot => 'boot',
    bootdisk => 'bootdisk',
    smp => 'natural',
    acpi => 'bool',
    kvm => 'bool',
    tdf => 'bool',
    localtime => 'bool',
    freeze => 'bool',
    vlanu => 'vlan',
    vga => 'vgatype',
    hostusb => 'hostusb',
    serial => 'serial',
    parallel => 'parallel',
    startdate => 'startdate',
    args => 'string',

    #cpu => 'string',
    #machine => 'string',
    #fda => 'file',
    #fdb => 'file',
    #mtdblock => 'file',
    #sd => 'file',
    #pflash => 'file',
    #snapshot => 'bool',
    #bootp => 'file',
    ##tftp => 'dir',
    ##smb => 'dir',
    ##usb
    ##usbdevice
    #kernel => 'file',
    #append => 'string',
    #initrd => 'file',

    ##soundhw => 'string',
};

my $MAX_IDE_DISKS = 4;
my $MAX_SCSI_DISKS = 16;
my $MAX_VIRTIO_DISKS = 16;

for (my $i = 0; $i < 10; $i++)  {
    $confvars->{"vlan$i"} = 'vlan';
}

my $drivename_hash;

for (my $i = 0; $i < $MAX_IDE_DISKS; $i++)  {
    $confvars->{"ide$i"} = 'drive';
    $drivename_hash->{"ide$i"} = 1;
}
for (my $i = 0; $i < $MAX_SCSI_DISKS; $i++)  {
    $confvars->{"scsi$i"} = 'drive';
    $drivename_hash->{"scsi$i"} = 1;
}
for (my $i = 0; $i < $MAX_VIRTIO_DISKS; $i++)  {
    $confvars->{"virtio$i"} = 'drive';
    $drivename_hash->{"virtio$i"} = 1;
}

my $defvars = {
    onboot => ['bool', 0],
    autostart => ['bool', 0],
    keyboard => ['lang', undef],
    cpuunits => ['natural', 1000],
    memory => ['natural', 512],
    vga => ['vgatype', 'cirrus'],
    tdf => ['bool', 1],
};

my $kvm_api_version = 0;

sub kvm_version {

    return $kvm_api_version if $kvm_api_version;

    my $fh = IO::File->new ("</dev/kvm") ||
	return 0;

    if (my $v = $fh->ioctl (KVM_GET_API_VERSION(), 0)) {
	$kvm_api_version = $v;
    }

    $fh->close();

    return  $kvm_api_version;
}

sub disknames {
    # order is important - used to autoselect boot disk
    return ((map { "ide$_" } (0 .. ($MAX_IDE_DISKS - 1))),  
	    (map { "scsi$_" } (0 .. ($MAX_SCSI_DISKS - 1))),
	    (map { "virtio$_" } (0 .. ($MAX_VIRTIO_DISKS - 1))));
}

sub valid_drivename {
    my $dev = shift;

    return defined ($drivename_hash->{$dev});
}

sub nic_models {
    return ['rtl8139', 'ne2k_pci', 'e1000',  'pcnet',  'virtio',
	    'ne2k_isa', 'i82551', 'i82557b', 'i82559er'];
}

sub os_list_description {

    return {
	other => 'Other',
	wxp => 'Windows XP',
	w2k => 'Windows 2000',
	w2k3 =>, 'Windows 2003',
	w2k8 => 'Windows 2008',
	wvista => 'Windows Vista',
	l24 => 'Linux 2.4',
	l26 => 'Linux 2.6',
    }; 
}

sub run_command {
    my ($cmd, $timeout, $outfunc) = @_;

    my $reader = IO::File->new();
    my $writer = IO::File->new();
    my $error  = IO::File->new();

    $cmd = [ $cmd ] if !ref($cmd);

    my $cmdstr = join (' ', @$cmd);

    my $orig_pid = $$;

    my $pid;
    eval {
	$pid = open3 ($writer, $reader, $error, @$cmd) || die $!;
    };

    my $err = $@;

    # catch exec errors
    if ($orig_pid != $$) {
	warn "ERROR: $err";
	POSIX::_exit (1); 
	kill ('KILL', $$); 
    }

    die $err if $err;

    close $writer;

    my $select = new IO::Select;
    $select->add ($reader);
    $select->add ($error);

    my $outlog = '';

    while ($select->count) {
	my @handles = $select->can_read ($timeout);

	if (defined ($timeout) && (scalar (@handles) == 0)) {
	    kill (9, $pid);
	    waitpid ($pid, 0);
	    die "command '$cmdstr' failed: timeout";
	}

	foreach my $h (@handles) {
	    my $buf = '';
	    my $count = sysread ($h, $buf, 4096);
	    if (!defined ($count)) {
		my $err = $!;
		kill (9, $pid);
		waitpid ($pid, 0);
		die "command '$cmdstr' failed: $err";
	    }
	    $select->remove ($h) if !$count;
	    if ($h eq $reader) {
		if ($outfunc) {
		    $outlog .= $buf;
		    while ($outlog =~ s/^([^\010\r\n]*)(\r|\n|(\010)+|\r\n)//s) {
			my $line = $1;
			&$outfunc($line) if $outfunc;
		    }
		} else {
		    print $buf;
		}
	    } elsif ($h eq $error) {
		print STDERR $buf;
	    }
	}
    }

    &$outfunc($outlog) if $outfunc && $outlog;

    my $rv = waitpid ($pid, 0);
    my $ec = ($? >> 8);

    if ($ec) {  
	die "command '$cmdstr' failed with exit code $ec";
    }
}

# a clumsy way to split an argument string into an array,
# we simply pass it to the cli (exec call)
sub split_args {
    my ($str) = @_;

    my $args = [];

    return $args if !$str;

    my $cmd = 'perl -e \'foreach my $a (@ARGV) { print "$a\n"; } \' -- ' . $str;

    eval {
	run_command ($cmd, undef, sub {
	    my $data = shift;
	    push @$args, $data;
	});
    };
    
    my $err = $@;

    die "unable to parse args: $str\n" if $err;

    return $args;
}

sub disk_devive_info {
    my $dev = shift;

    die "unknown disk device format '$dev'" if $dev !~ m/^(ide|scsi|virtio)(\d+)$/;

    my $bus = $1;
    my $index = $2;
    my $maxdev = 1024;

    if ($bus eq 'ide') {
	$maxdev = 2;
    } elsif ($bus eq 'scsi') {
	$maxdev = 8;
    }

    my $controller = int ($index / $maxdev);
    my $unit = $index % $maxdev;


    return { bus => $bus, desc => uc($bus) . " $controller:$unit",
	     controller => $controller, unit => $unit, index => $index };

}

sub qemu_drive_name {
    my ($dev, $media) = @_; 

    my $info = disk_devive_info ($dev);
    my $mediastr = '';

    if (($info->{bus} eq 'ide') || ($info->{bus} eq 'scsi')) {
	$mediastr = ($media eq 'cdrom') ? "-cd" : "-hd";
	return sprintf("%s%i%s%i", $info->{bus}, $info->{controller}, 
		       $mediastr, $info->{unit});
    } else {
	return sprintf("%s%i", $info->{bus}, $info->{index}); 
    }
}
 
my $imagedir_abs;
my $isodir_abs;

sub relative_path {
    my ($vmid, $media, $path) = @_;

    return $path if $path !~ m|/|;
    return $path if $path =~ m|^/dev/.+|;

    $imagedir_abs = abs_path ($imagedir) if !$imagedir_abs;
    $isodir_abs = abs_path ($isodir) if !$isodir_abs;

    my $abspath =  abs_path ($path);
    if (!$abspath) {
	warn "image '$path' does not exists\n";
	return undef;
    }

    my $imagepath;

    if ($media eq 'disk') {
	$imagepath = "$imagedir_abs/$vmid/";
    } elsif ($media eq 'cdrom') {
	$imagepath = "$isodir_abs/";
    } else {
	die "unknowm media '$media'";
    }

    $abspath =~ s/^$imagepath//;

    if ($abspath =~ m|/|) {
	warn "image '$path' outside image directory\n";
	return undef;
    }
    return $abspath;
}

my $cdrom_path;

sub get_cdrom_path {

    return  $cdrom_path if $cdrom_path;

    return $cdrom_path = "/dev/cdrom" if -l "/dev/cdrom";
    return $cdrom_path = "/dev/cdrom1" if -l "/dev/cdrom1";
    return $cdrom_path = "/dev/cdrom2" if -l "/dev/cdrom2";
}

sub get_iso_path {
    my ($vmdi, $cdrom) = @_;

    if ($cdrom eq 'cdrom') {
	return get_cdrom_path();
    } else {
	$cdrom = "$isodir/$cdrom" if substr ($cdrom, 0, 1) ne '/';
    }

    return $cdrom;
}

sub parse_options {
    my ($vmid, $create) = @_;

    my $settings = {};

    my $get_option_func = sub {
	my ($opt, $value) = @_;
	    
	if ($opt eq 'cdrom') {
	    $opt = 'ide2';
	    if ($value && $value ne 'undef' && $value ne 'delete' && 
		$value !~ m/media=/) {
		$value .= ",media=cdrom"; 
	    }
	}

	my $ct = $confvars->{$opt};

	my $v;
	if (defined ($v = check_type ($vmid, $ct, $opt, $value, 1))) {
	    $settings->{$opt} = $v;
	} else {
	    if (!$create && ($value eq '' || $value eq 'undef')) {
		$settings->{$opt} = $value;
	    } else {
		die "unable to parse value for '$opt'\n";
	    }
	}
    };

    my $getopt;
    foreach my $opt (keys %$confvars) {
	$getopt->{"$opt=s"} = $get_option_func;
    }

    $getopt->{"cdrom=s"} = $get_option_func;
 
    if (!GetOptions (%$getopt)) {
	return undef;
    }

    return $settings;
}

sub create_conf {
    my ($vmid, $settings) = @_;

    lock_config ($vmid, &create_conf_nolock, $settings);
}

sub create_conf_nolock {
    my ($vmid, $settings) = @_;

    my $filename = config_file ($vmid);

    die "configuration file '$filename' already exists\n" if -f $filename;
    
    my $defaults = load_defaults();

    $settings->{name} = "vm$vmid" if !$settings->{name};
    if (!$settings->{memory}) {
	$settings->{memory} = $defaults->{memory} || $default_memory;
    }

    my $fh = new IO::File ($filename, "w") ||
	die "unable to create config for VM $vmid - $!\n";
    
    foreach my $opt (keys %$settings) {
	next if !$confvars->{$opt};

	my $value = $settings->{$opt};
	next if !$value;

	print $fh "$opt: $value\n";
    }

    $fh->close();
}

sub image_dir {
    my ($vmid) = @_;

    return "$imagedir/$vmid";
}

sub image_path {
    my ($vmid, $filename) = @_;

    my $path = $filename;
    $path = "$imagedir/$vmid/$path" if substr ($path, 0, 1) ne '/';
    return $path;
}

sub create_image {
    my ($vmid, $size, $fmt) = @_; # size in GB

    my $sizemb = int ($size*1024);

    mkpath "$imagedir/$vmid";

    my $filename = "vm-$vmid-disk.$fmt";
    my $path = image_path ($vmid, $filename);;
	
    if (-f $path) {
	for (my $i = 1; $i < 100; $i++) {
	    $filename = "vm-$vmid-disk-$i.$fmt";
	    $path = image_path ($vmid, $filename);
	    last if ! -f $path;
	}
    }

    die "disk image '$path' already exists\n" if -f $path;

    if (system ("/usr/bin/qemu-img create -f $fmt '$path' ${sizemb}M") != 0) {
	my $msg = "unable to create image - $!\n";
	unlink $path;
	die $msg;
    }

    return $filename;
}

# ideX = [file=]filename[,media=d][,cyls=c,heads=h,secs=s[,trans=t]]
#        [,snapshot=on|off][,cache=on|off][,format=f]

sub parse_drive {
    my ($key, $data, $cli) = @_;

    my $res = {};
    
    if ($key =~ m/^([^\d]+)(\d+)$/) {
	$res->{interface} = $1;
	$res->{index} = $2;
    } else {
	return undef;
    }

    foreach my $p (split (/,/, $data)) {
	next if $p =~ m/^\s*$/;

	if ($p =~ m/^(file|cyls|heads|secs|trans|media|snapshot|cache|format)=(.+)$/) {
	    return undef if defined $res->{$1};
	    $res->{$1} = $2;
	} else {
	    if (!$res->{file} && $p !~ m/=/) {
		$res->{file} = $p;
	    } else {
		return undef;
	    }
	}
    }

    return undef if !$res->{file};

    if (!$cli) {
	if (substr ($res->{file}, 0, 1) eq '/') {
	    if ($res->{file} !~ m|^/dev/.+|) {
		return undef;
	    }
	}
    }

    return undef if $res->{cache} && $res->{cache} !~ m/^(on|off)$/;
    return undef if $res->{snapshot} && $res->{snapshot} !~ m/^(on|off)$/;
    return undef if $res->{cyls} && $res->{cyls} !~ m/^\d+$/;
    return undef if $res->{heads} && $res->{heads} !~ m/^\d+$/;
    return undef if $res->{secs} && $res->{secs} !~ m/^\d+$/;
    return undef if $res->{media} && $res->{media} !~ m/^(disk|cdrom)$/;
    return undef if $res->{trans} && $res->{trans} !~ m/^(none|lba|auto)$/;
    return undef if $res->{format} && $res->{format} !~ m/^(raw|cow|qcow|qcow2|vmdk|cloop)$/;

    if ($res->{media} && ($res->{media} eq 'cdrom')) {
	return undef if $res->{snapshot} || $res->{trans} || $res->{format};
	return undef if $res->{heads} || $res->{secs} || $res->{cyls}; 
	return undef if $res->{interface} eq 'virtio';
    }

    return $res;
}

sub print_drive {
    my ($vmid, $drive, $full) = @_;

    my $res = '';

    if ($full) {
	my $path;
	
	if (drive_is_cdrom ($drive)) {
	    if ($drive->{file} eq 'none') {
		$path = '';
	    } else {
		$path = get_iso_path ($vmid, $drive->{file});
	    }
	} else {
	    $path = image_path ($vmid, $drive->{file});
	}
	$res = "file=$path,if=$drive->{interface},index=$drive->{index}";
    } else {
	$res = "$drive->{file}";
    }
    foreach my $o (qw(heads secs cyls trans media format cache snapshot)) {
	$res .= ",$o=$drive->{$o}" if $drive->{$o};
    }

    return $res;
}

sub drive_is_cdrom {
    my ($drive) = @_;

    return $drive && $drive->{media} && ($drive->{media} eq 'cdrom');

}
# vlanX: e1000=XX:XX:XX:XX:XX:XX,pcnet=XX:XX:XX:XX:XX:XX,...

sub parse_vlan {
    my ($data) = @_;

    my $res = {};

    foreach my $nic (split (/,/, $data)) {

	if ($nic =~ m/^(ne2k_pci|e1000|rtl8139|pcnet|virtio|ne2k_isa|i82551|i82557b|i82559er)(=([0-9a-f]{2}(:[0-9a-f]{2}){5}))?$/i) {
	    my $model = lc ($1);
	    my $mac = uc($3) || random_ether_addr ();
	    push @{$res->{nics}}, { model => $model, macaddr => $mac } ;
	} else {
	    return undef;
	}
	
    }

    return undef if !$res->{nics};

    return $res;

}

sub print_vlan {
    my $vlan = shift;

    my $res = '';
    foreach my $nic (@{$vlan->{nics}}) {
	$res .= ',' if $res;
	$res .= "$nic->{model}";
	my $mac =  $nic->{macaddr};
	$res .= "=$mac" if $mac;
    }
    return $res;
}

sub check_type {
    my ($vmid, $type, $key, $value, $cli) = @_;

    my $res;

    if ($type eq 'bool') {
	return 1 if ($value eq '1') || ($value =~ m/^(on|yes|true)$/i); 
	return 0 if ($value eq '0') || ($value =~ m/^(off|no|false)$/i); 
    } elsif ($type eq 'natural') {
	return $1 if $value =~ m/^(\d+)$/;
    } elsif ($type eq 'lang') {
	$value = lc ($value);
	return $value if -f "/usr/share/kvm/keymaps/$value";
    } elsif ($type eq 'file') {
	$value =~ s/^\"(.*)\"$/$1/;
	return $value;	
    } elsif ($type eq 'string') {
	$value =~ s/^\"(.*)\"$/$1/;
	return $value;	
    } elsif ($type eq 'ostype') {
	return $value if $value =~ m/^(other|wxp|w2k|w2k3|w2k8|wvista|l24|l26)$/;
    } elsif ($type eq 'vgatype') {
	return $value if $value =~ m/^(std|cirrus)$/;
    } elsif ($type eq 'boot') {
	return $value if $value =~ m/^[acdn]{1,3}$/;
    } elsif ($type eq 'bootdisk') {
	foreach my $ds (disknames()) {
	    return $value if $value eq $ds;
	}
	return undef;
    } elsif ($type eq 'vlan') {
	return $value if parse_vlan ($value);	
    } elsif ($type eq 'drive') {
	# when invoked with $cli, compute relative path
	my $drive = parse_drive ($key, $value, $cli);
	if ($drive && $cli) {
	    my $media = $drive->{media} || 'disk';
	    if (my $file = relative_path ($vmid, $media, $drive->{file})) {
		$drive->{file} = $file;
		return print_drive ($vmid, $drive);
	    }
	    return undef;
	} else {
	    return $value if $drive;
	}	
    } elsif ($type eq 'hostusb') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m/^(\d+.(\d+|\*)|[0-9A-Fa-f]+:([0-9A-Fa-f]+|\*))$/
	}
	return $value;
    } elsif ($type eq 'parallel') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m|^/dev/parport\d+$|
	}
	return $value;	
    } elsif ($type eq 'serial') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m|^/dev/ttyS\d+$|
	}
	return $value;	
    } elsif ($type eq 'startdate') {
	return $value if $value eq 'now';
	return $value if $value =~ m/^\d{4}-\d{1,2}-\d{1,2}(T\d{1,2}-\d{1,2}-\d{1,2})?$/;
    }

    return undef;
}

sub lock_config {
    my ($vmid, $code, @param) = @_;

    my $filename = config_file_lock ($vmid);

    lock_file ($filename, "VM $vmid", $code, @param);
}

sub lock_file {
    my ($filename, $text, $code, @param) = @_;

    my $timeout = 10;

    my $fh;

    my $res;

    eval {

	local $SIG{ALRM} = sub { die "got timeout\n"; };

	alarm ($timeout);

	$fh = new IO::File (">>$filename") ||
	    die "can't open lock for $text '$filename' - $!\n";

	if (!flock ($fh, LOCK_EX|LOCK_NB)) {
	    print STDERR "trying to aquire lock...";
	    if (!flock ($fh, LOCK_EX)) {
		print STDERR " failed\n";
		die "can't aquire lock for $text '$filename' - $!\n";
	    }
	    print STDERR " OK\n";
	}
	alarm (0);

	$res = &$code (@param);
    };

    my $err = $@;

    alarm (0);

    close ($fh) if $fh;

    if ($err) {
	$@ = $err;
	return undef;
    }

    return $res;
}

sub config_file {
    my ($vmid) = @_;

    return "$confdir/$vmid.conf";
}

sub config_file_lock {
    my ($vmid) = @_;

    return "$confdir/.lock-$vmid.conf";
}

sub touch_config {
    my ($vmid) = @_;

    my $conf = config_file ($vmid);
    utime undef, undef, $conf;
}

sub unlink_image {
    my ($vmid, $file) = @_;

   # only unlink disks inside $imagedir
    return if $file =~ m|/|;
    if (my $path = image_path ($vmid, $file)) {
	unlink $path if -f $path;
	touch_config ($vmid);
    }
}

sub destroy_vm {
    my ($vmid) = @_;

    my $conffile = config_file ($vmid);

    my $conf = load_config ($vmid);

    my $di =  $conf->{diskinfo};

    my $cmd = "(rm -f '$conffile'";

    # only remove disks inside $imagedir
    foreach my $ds (keys %$di) {
	next if $ds =~ m/^unused/;
	next if drive_is_cdrom ($di->{$ds});

	my $file = $di->{$ds}->{file};
	next if !$file;
	next if $file =~ m|/|;
	if (my $path = image_path ($vmid, $file)) {
	    if (-f $path) {
		$cmd .= " '$path'";
	    }
	}
    }
    
    $cmd .= ";rmdir $imagedir/$vmid)";

    system ("$cmd &"); # run in background
}

sub load_disksize {
    my ($vmid, $file) = @_;

    my $path = image_path ($vmid, $file);

    my $data = `/usr/bin/qemu-img info '$path'`;

    if ($data =~ m/^virtual size:\s\S+\s+\((\d+)\s+bytes\)$/mi) {
	return $1/(1024*1024*1024);
    }

    return 0;
}

sub load_diskinfo {
    my ($vmid, $conf) = @_;

    my @disks = disknames();

    my $unused = {};
    foreach my $file (<$imagedir/$vmid/*>) {
	next if ! -f $file;
	$file = basename ($file);
	next if $file !~ m/^.+\.(qcow2|raw|vmdk|qcow|cow)$/;
	$unused->{$file} = 1;
    }

    my $res = {};
    foreach my $ds (@disks) {
	next if !$conf->{$ds};

	my $di = parse_drive ($ds, $conf->{$ds});

	next if !$di;

	$res->{$ds} = $di;

	delete ($unused->{$di->{file}});

	next if drive_is_cdrom ($di);

	$res->{$ds}->{disksize} = load_disksize ($vmid, $di->{file});
    }

    my $i = 0;
    foreach my $file (keys %$unused) {
	if (my $size = load_disksize ($vmid, $file)) {
	    $res->{"unused$i"} = { file => $file, disksize => $size };
	    $i++;
	}
    }

    return $res;
}

sub load_config {
    my ($vmid) = @_;

    my $filename = config_file ($vmid);

    my $fh = new IO::File ($filename, "r") ||
	die "unable to read config for VM $vmid\n";
 

    return parse_config ($filename, $fh);
}  

sub parse_config {
    my ($filename, $fh) = @_;

    my $res = {};

    $filename =~ m|/(\d+)\.conf$| 
	|| die "got strange filename '$filename'";

    my $vmid = $1;

    my $imagestore = "$imagedir/$vmid";

    while (my $line = <$fh>) {
	
	next if $line =~ m/^\#/;

	next if $line =~ m/^\s*$/;

	if ($line =~ m/^(description|args):\s*(.*\S)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    $res->{$key} = $value;
	} elsif ($line =~ m/^([a-z]+\d*):\s*(\S+)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    if (my $type = $confvars->{$key}) {
		if (defined ($value = check_type ($vmid, $type, $key, $value))) {
		    $res->{$key} = $value;
		} else {
		    warn "vm $vmid - unable to parse value of '$key'\n";
		}
	    } else {
		warn "vm $vmid - unknown setting '$key'\n";
	    }
	}
    }

    my $disksize = 0;
    my $disktype = 'none';
 
    my $di = load_diskinfo ($vmid, $res);

    foreach my $ds (keys %$di) {
	my $size = $di->{$ds}->{disksize};
	if ($res->{bootdisk} && ($ds eq $res->{bootdisk}) && $size) {
	    $disksize = $size;
	    $disktype = $di->{$ds}->{interface};
	}
    }

    $res->{disksize} = $disksize;
    $res->{disktype} = $disktype;
    $res->{diskinfo} = $di;

    return $res;
}


sub change_config {
    my ($vmid, $settings, $unset) = @_;

    lock_config ($vmid, &change_config_nolock, $settings, $unset);
}

sub change_config_nolock {
    my ($vmid, $settings, $unset) = @_;

    my $res = {};

    foreach my $key (keys %$settings) {
	my $value = $settings->{$key};
	if (my $type = $confvars->{$key}) {
	    if (defined ($value = check_type ($vmid, $type, $key, $value))) {
		$res->{$key} = $value;
	    } else {
		die "unable to parse value of '$key'\n";
	    }
	} else {
	    die "unknown setting '$key'\n";
	}
    }

    my $filename = config_file ($vmid);
    my $tmpfn = "$filename.$$.tmp";

    my $fh = new IO::File ($filename, "r") ||
	die "unable to read config for VM $vmid\n";

    my $out = new IO::File ($tmpfn, "w") ||
	die "unable to write config for VM $vmid\n";

    eval {

	while (my $line = <$fh>) {
	
	    if (($line =~ m/^\#/) || ($line =~ m/^\s*$/)) {
		print $out $line;
		next;
	    }

	    if ($line =~ m/^([a-z]+\d*):\s*(.*\S)\s*$/) {
		my $key = $1;
		my $value = $2;
		if (defined ($res->{$key})) {
		    $value = $res->{$key};
		    delete $res->{$key};
		}
		print $out "$key: $value\n" if !defined ($unset->{$key});
		next;
	    }

	    die "unable to parse config file: $line\n";
	}

	foreach my $key (keys %$res) {
	    print $out "$key: $res->{$key}\n" if !defined ($unset->{$key});	    
	}
    };

    my $err = $@;

    if ($err) {
	unlink $tmpfn;
	die $err;
    } else {
	rename $tmpfn, $filename;
    }
}

sub load_defaults { 

    my $filename = "/etc/pve/qemu-server.cfg";

    my $res = {};

    while (my ($key, $value) = each %$defvars) {
	my ($type, $default) = @$value;
	$res->{$key} = $default if defined ($default);
    }
   
    my $fh = new IO::File ($filename, "r") || return $res;

    while (my $line = <$fh>) {
	
	next if $line =~ m/^\#/;

	next if $line =~ m/^\s*$/;

	if ($line =~ m/^([a-z]+):\s*(\S+)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    my ($type, $default);

	    ($type, $default) = @{$defvars->{$key}} if defined $defvars->{$key};

	    if ($type) {
		if (defined ($value = check_type (undef, $type, $key, $value))) {
		    $res->{$key} = $value;
		} else {
		    warn "load_defaults - unable to parse value of '$key'\n";
		}
	    } else {
		warn "load_defaults - unknown setting '$key'\n";
	    }
	}
    }

    return $res;
}

sub config_list {

    my $fd = IO::Dir->new ($confdir);
    my $res = {};

    return $res if !$fd;

    while (defined(my $de = $fd->read)) { 
	if ($de =~ m/^(\d+)\.conf$/) {
	    $res->{$1}->{exists} = 1;
	}
    }
    return $res;
}

sub check_cmdline {
    my ($pidfile, $pid) = @_;

    my $fh = IO::File->new ("/proc/$pid/cmdline", "r");
    if (defined ($fh)) {
	my $line = <$fh>;
	$fh->close;
	return undef if !$line;
	my @param = split (/\0/, $line);

	my $cmd = $param[0];
	return if !$cmd || ($cmd !~ m|kvm$|);

	for (my $i = 0; $i < scalar (@param); $i++) {
	    my $p = $param[$i];
	    next if !$p;
	    if (($p eq '-pidfile') || ($p eq '--pidfile')) {
		my $p = $param[$i+1];
		return 1 if $p && ($p eq $pidfile);
		return undef;
	    }
	}
    }
    return undef;
}

sub check_running {
    my ($vmid) = @_;

    my $filename = config_file ($vmid);

    die "unable to find configuration file for VM $vmid - no such machine\n"
	if ! -f $filename;

    my $pidfile = pidfile_name ($vmid);

    if (my $fd = IO::File->new ("<$pidfile")) {
	my $st = stat ($fd);
	my $line = <$fd>;
	close ($fd);

	my $mtime = $st->mtime;
	if ($mtime > time()) {
	    warn "file '$filename' modified in future\n";
	}

	if ($line =~ m/^(\d+)$/) {
	    my $pid = $1;

	    return $pid if ((-d "/proc/$pid") && check_cmdline ($pidfile, $pid));
	}
    }

    return undef;
}

sub vzlist {
    
    my $vzlist = config_list();

    my $fd = IO::Dir->new ($var_run_tmpdir) || return $vzlist;

    while (defined(my $de = $fd->read)) { 
	next if $de !~ m/^(\d+)\.pid$/;
	my $vmid = $1;
	next if !defined ($vzlist->{$vmid});
	if (my $pid = check_running ($vmid)) {
	    $vzlist->{$vmid}->{pid} = $pid;
	}
    }

    return $vzlist;
}

sub config_to_command {
    my ($vmid, $conf, $statefile) = @_;

    my $defaults = load_defaults ();

    my $cmd = [];

    push @$cmd, '/usr/bin/kvm';

    my $socket = monitor_socket ($vmid);
    push @$cmd, '-monitor', "unix:$socket,server,nowait";

    $socket = vnc_socket ($vmid);
    push @$cmd,  '-vnc', "unix:$socket,password";

    push @$cmd, '-pidfile' , pidfile_name ($vmid);
 
    push @$cmd, '-daemonize';

    push @$cmd, '-incoming', "exec:cat $statefile" if $statefile && -f $statefile;

    # enable absolute mouse coordinates (needed by vnc)
    push @$cmd, '-usbdevice', 'tablet'; 

    # host usb devices
    if (my $usbdl = $conf->{hostusb}) {
	my @dl = split (/,/, $usbdl);
	foreach my $dev (@dl) {
	    push @$cmd, '-usbdevice', "host:$dev" if $dev;
	}
    }

    # serial devices
    if (my $serdl = $conf->{serial}) {
	my @dl = split (/,/, $serdl);
	foreach my $dev (@dl) {
	    next if !$dev;
	    if (-c $dev) {
		push @$cmd, '-serial', "$dev";
	    }
	}
    }

    # parallel devices
    if (my $pardl = $conf->{parallel}) {
	my @dl = split (/,/, $pardl);
	foreach my $dev (@dl) {
	    next if !$dev;
	    if (-c $dev) {
		push @$cmd, '-parallel', "$dev";
	    }
	}
    }

    my $vmname = $conf->{name} || "vm$vmid";

    push @$cmd, '-name', $vmname;
    
    push @$cmd, '-smp', $conf->{smp} if $conf->{smp};

    push @$cmd, '-id', $vmid;
 
    push @$cmd, '-cpuunits', $conf->{cpuunits} || 1000;

    # fixme: cpulimit is currently ignored
    #push @$cmd, '-cpulimit', $conf->{cpulimit} if $conf->{cpulimit};
    
    push @$cmd, '-boot', $conf->{boot} if $conf->{boot};

    push @$cmd, '-no-acpi' if defined ($conf->{acpi}) && $conf->{acpi} == 0;

    push @$cmd, '-no-reboot' if  defined ($conf->{reboot}) && $conf->{reboot} == 0;

    my $vga = $conf->{vga} || $defaults->{vga};
    push @$cmd, '-vga', $vga if $vga; # for kvm 77 and later

    # time drift fix
    my $tdf = defined ($conf->{tdf}) ? $conf->{tdf} : $defaults->{tdf};
    push @$cmd, '-tdf' if $tdf;

    my $nokvm = defined ($conf->{kvm}) && $conf->{kvm} == 0 ? 1 : 0;

    if (my $ost = $conf->{ostype}) {
	# other, wxp, w2k, w2k3, w2k8, wvista, l24, l26

	if ($ost =~ m/^w/) { # windows
	    push @$cmd, '-localtime' if !defined ($conf->{localtime}); 
	}

	# -tdf ?
	# -no-acpi 
	# -no-kvm 
	# -win2k-hack ?
    }

    push @$cmd, '-no-kvm' if $nokvm;

    push @$cmd, '-localtime' if $conf->{localtime};

    push @$cmd, '-startdate', $conf->{startdate} if $conf->{startdate};

    push @$cmd, '-S' if $conf->{freeze};

    # set keyboard layout
    my $kb = $conf->{keyboard} || $defaults->{keyboard};
    push @$cmd, '-k', $kb if $kb;

    # enable sound
    #my $soundhw = $conf->{soundhw} || $defaults->{soundhw};
    #push @$cmd, '-soundhw', 'es1370';
    #push @$cmd, '-soundhw', $soundhw if $soundhw;

    my $di = $conf->{diskinfo};

    foreach my $ds (keys %$di) {
	next if $ds =~ m/^unused/;
	my $drive = print_drive ($vmid, $di->{$ds}, 1);
	$drive .= ",boot=on" if $conf->{bootdisk} && ($conf->{bootdisk} eq $ds);
	push @$cmd, '-drive', $drive;
    }

    if ($conf->{memory}) {
	push @$cmd, '-m', $conf->{memory};
    } else {
	push @$cmd, '-m', $defaults->{memory} || $default_memory;
    }

    my $foundnet = 0;

    for (my $i = 0; $i < 10; $i++) {

	if ($conf->{"vlan$i"} && (my $vlan = parse_vlan ($conf->{"vlan$i"}))) {

	    $foundnet = 1;

	    my $ifname = "vmtab${vmid}i$i";
	    push @$cmd, '-net', "tap,vlan=$i,ifname=$ifname,script=/var/lib/qemu-server/bridge-vlan$i";

	    foreach my $nic (@{$vlan->{nics}}) {
		push @$cmd, '-net', "nic,vlan=$i,model=$nic->{model},macaddr=$nic->{macaddr}";
	    }

	}
    } 

    if ($conf->{"vlanu"} && (my $vlan = parse_vlan ($conf->{"vlanu"}))) {

	$foundnet = 1;

	push @$cmd, '-net', "user,vlan=10,hostname=$vmname";

	foreach my $nic (@{$vlan->{nics}}) {
	    push @$cmd, '-net', "nic,vlan=10,model=$nic->{model},macaddr=$nic->{macaddr}";
	}
    }
    
    push @$cmd, '-net', 'none' if !$foundnet;

    # add custom args
    if ($conf->{args}) {
	my $aa = split_args ($conf->{args});
	push @$cmd, @$aa;
    }

    return $cmd;
}
 
sub vnc_socket {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.vnc";
}

sub monitor_socket {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.mon";
}

sub pidfile_name {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.pid";
}

sub random_ether_addr {

    my $rand = Digest::SHA1::sha1_hex (rand(), time());

    my $mac = '';
    for (my $i = 0; $i < 6; $i++) {
	my $ss = hex (substr ($rand, $i*2, 2));
	if (!$i) {
	    $ss &= 0xfe; # clear multicast
	    $ss |= 2; # set local id
	}
	$ss = sprintf ("%02X", $ss);

	if (!$i) {
	    $mac .= "$ss";
	} else {
	    $mac .= ":$ss";
	}
    }
    
    return $mac;
}


sub vm_start {
    my ($vmid, $statefile) = @_;

    my $conf = load_config ($vmid);

    mkpath $var_run_tmpdir;
    
    if (check_running ($vmid)) {
	my $msg = "VM $vmid already running - start failed\n" ;
	syslog ('err', $msg);
	die $msg;
    } else {
	syslog ('info', "VM $vmid start");
    }

    my $cmd = PVE::QemuServer::config_to_command ($vmid, $conf, $statefile);
    
    eval  { run_command ($cmd, $statefile ? undef : 30); };

    my $err = $@;

    if ($err) {
	my $msg = "start failed: $err";
	syslog ('err', "VM $vmid $msg");
	die $msg;
    }

    if ($statefile) {
	unlink $statefile;
	# send resume
	eval  { vm_monitor_command ($vmid, "cont", 1) };
	syslog ('err', "VM $vmid resume failed - $@") if $@;
    }
}

sub __read_avail {
    my ($fh, $timeout) = @_;

    my $sel = new IO::Select;
    $sel->add ($fh);

    my $res = '';
    my $buf;

    my @ready;
    while (scalar (@ready = $sel->can_read ($timeout))) {
	my $count;
	if ($count = $fh->sysread ($buf, 8192)) {
	    if ($buf =~ /^(.*)\(qemu\) $/s) {
		$res .= $1;
		last;
	    } else {
		$res .= $buf;
	    }
	} else {
	    if (!defined ($count)) {
		die "$!\n";
	    }
	    last;
	}
    }

    die "monitor read timeout\n" if !scalar (@ready);

    return $res;
}

sub vm_monitor_command {
    my ($vmid, $cmdstr, $nolog) = @_;

    my $res;

    syslog ("info", "VM $vmid monitor command '$cmdstr'") if !$nolog;

    eval {
	die "VM not running\n"  if !check_running ($vmid);

	my $sname = monitor_socket ($vmid);

	my $sock = IO::Socket::UNIX->new ( Peer => $sname ) ||
	    die "unable to connect to VM $vmid socket - $!\n";

	my $timeout = 3;

	# read banner;
	my $data = __read_avail ($sock, $timeout);
	
	if ($data !~ m/^QEMU\s+(\S+)\s+monitor\s/) {
	    die "got unexpected qemu monitor banner\n";
	}

	my $sel = new IO::Select;
	$sel->add ($sock);

	if (!scalar (my @ready = $sel->can_write ($timeout))) {
	    die "monitor write error - timeout";
	}

	my $fullcmd = "$cmdstr\r";

	my $b;
	if (!($b = $sock->syswrite ($fullcmd)) || ($b != length ($fullcmd))) {
	    die "monitor write error - $!";
	}

	return if ($cmdstr eq 'q') || ($cmdstr eq 'quit');

	$timeout = 20; 

	if ($cmdstr =~ m/^migrate/) {
	    $timeout = 60*60; # 1 hour
	} elsif ($cmdstr =~ m/^(eject|change)/) {
	    $timeout = 60; # note: cdrom mount command is slow
	}

	if ($res = __read_avail ($sock, $timeout)) {
  
	    my @lines = split ("\r?\n", $res);

	    shift @lines if $lines[0] !~ m/^unknown command/; # skip echo
	    
	    $res = join ("\n", @lines);
	    $res .= "\n";
	}
    };

    my $err = $@;

    if ($err) {
	syslog ("err", "VM $vmid monitor command failed - $err");
	die $err;
    }

    return $res;
}

sub vm_vncticket {
    my ($vmid, $ticket) = @_;

    # generate ticket, olny first 8 character used by vnc
    $ticket = Digest::SHA1::sha1_base64 (rand(), time()) if !$ticket;

    vm_monitor_command ($vmid, "change vnc ticket $ticket", 1);

    return $ticket;
}

sub vm_commandline {
    my ($vmid) = @_;

    my $conf = PVE::QemuServer::load_config ($vmid);

    my $cmd = PVE::QemuServer::config_to_command ($vmid, $conf);

    return join (' ', @$cmd);
}

sub vm_reset {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid sending 'reset'");

    vm_monitor_command ($vmid, "system_reset", 1);
}

sub vm_wait {
    my ($vmid, $timeout, $nolog) = @_;

    my $pid = check_running ($vmid);

    syslog ("info", "VM $vmid waiting until stopped") if !$nolog;

    my $count = 0;
    while ((!$timeout || ($count < $timeout)) && check_running ($vmid)) {
	$count++;
	sleep 1;
    }

    if (check_running ($vmid)) {
	syslog ('info', "VM $vmid wait failed - got timeout");
	die "wait failed - got timeout\n";
    }
}

sub vm_shutdown {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid sending 'shutdown'");

    vm_monitor_command ($vmid, "system_powerdown", 1);
}

sub vm_stop {
    my ($vmid) = @_;

    my $pid = check_running ($vmid);

    if (!$pid) {
	syslog ('info', "VM $vmid already stopped");
	return;
    }
    
    syslog ("info", "VM $vmid stopping");

    eval { vm_monitor_command ($vmid, "quit", 1); };

    my $err = $@;

    if (!$err) {
	# wait some time
	my $timeout = 50; # fixme: how long?

	my $count = 0;
	while (($count < $timeout) && check_running ($vmid)) {
	    $count++;
	    sleep 1;
	}

	if ($count >= $timeout) {
	    syslog ('info', "VM $vmid still running - terminating now with SIGTERM");
	    kill 15, $pid;
	}
    } else {
	syslog ('info', "VM $vmid quit failed - terminating now with SIGTERM");
	kill 15, $pid;
    }

    # wait again
    my $timeout = 10;

    my $count = 0;
    while (($count < $timeout) && check_running ($vmid)) {
	$count++;
	sleep 1;
    }

    if ($count >= $timeout) {
	syslog ('info', "VM $vmid still running - terminating now with SIGKILL\n");
	kill 9, $pid;
    }

    fairsched_rmnod ($vmid); # try to destroy group
}

sub vm_suspend {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid suspend");

    vm_monitor_command ($vmid, "stop", 1);
}

sub vm_resume {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid resume");

    vm_monitor_command ($vmid, "cont", 1);
}

sub vm_cad {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid sending cntl-alt-delete");
    
    vm_monitor_command ($vmid, "sendkey ctrl-alt-delete", 1);
}

sub vm_cdrom {
    my ($vmid, $dev, $path) = @_;

    return if !check_running ($vmid); # do nothing

    my $qdn;
    eval {
	if (!valid_drivename ($dev)) {
	    die "cdrom change failed - invalid device name '$dev'\n";
	}

	if ($path eq 'eject' || $path eq 'none') {
	    syslog ("info", "VM $vmid eject cdrom '$dev'");
	} else {
	    syslog ("info", "VM $vmid change cdrom '$dev' to '$path'");
	}

	$qdn = qemu_drive_name ($dev, 'cdrom');
    };

    if (my $err = $@) {
	syslog ('err', "VM $vmid $err");
	die $err;
    }

    if ($path eq 'eject' || $path eq 'none') {
	vm_monitor_command ($vmid, "eject $qdn", 1);
    } else {
	$path = get_iso_path ($vmid, $path);

	vm_monitor_command ($vmid, "eject $qdn", 1);
	vm_monitor_command ($vmid, "change $qdn \"$path\"", 1);
    }
}

sub vm_destroy {
    my ($vmid) = @_;

    syslog ("info", "VM $vmid destroy called (removing all data)");

    eval {
	if (!check_running($vmid)) {
	    fairsched_rmnod ($vmid); # try to destroy group
	    destroy_vm ($vmid);
	} else {
	    die "VM is running\n";
	}
    };

    my $err = $@;

    if ($err) {
	syslog ("err", "VM $vmid destroy failed - $err");
	die $err;
    }
}

sub vm_startall {

    my $vzlist =  PVE::QemuServer::vzlist();

    foreach my $vmid (keys %$vzlist) {
	next if $vzlist->{$vmid}->{pid}; # already running

	eval {
	    my $conf =  PVE::QemuServer::load_config ($vmid);
	    if ($conf->{onboot}) {
		print STDERR "Starting Qemu VM $vmid\n";
		vm_start ($vmid);
	    }
	};
	print STDERR $@ if $@;
    }
}

sub vm_stopall {
    my $timeout = shift;

    $timeout = 3*60 if !$timeout;

    my $vzlist =  PVE::QemuServer::vzlist();
    my $count = 0;
    foreach my $vmid (keys %$vzlist) {
	next if !$vzlist->{$vmid}->{pid};
	$count++;
    }

    if ($count) {

	my $msg = "Stopping Qemu Server - sending shutdown requests to all VMs\n";
	syslog ('info', $msg);
	print STDERR $msg;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	    eval { vm_shutdown ($vmid); };
	    print STDERR $@ if $@;
	}

	my $wt = 5;
	my $maxtries = int (($timeout + $wt -1)/$wt);
	my $try = 0;
	while (($try < $maxtries) && $count) {
	    $try++;
	    sleep $wt;

	    $vzlist =  PVE::QemuServer::vzlist();
	    $count = 0;
	    foreach my $vmid (keys %$vzlist) {
		next if !$vzlist->{$vmid}->{pid};
		$count++;
	    }
	    last if !$count;
	}

	return if !$count;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	
	    $msg = "VM $vmid still running - sending stop now\n";
	    syslog ('info', $msg);
	    print $msg;

	    eval { vm_monitor_command ($vmid, "quit", 1); };
	    print STDERR $@ if $@;

	}

	$timeout = 30;
	$maxtries = int (($timeout + $wt -1)/$wt);
	$try = 0;
	while (($try < $maxtries) && $count) {
	    $try++;
	    sleep $wt;

	    $vzlist =  PVE::QemuServer::vzlist();
	    $count = 0;
	    foreach my $vmid (keys %$vzlist) {
		next if !$vzlist->{$vmid}->{pid};
		$count++;
	    }
	    last if !$count;
	}

	return if !$count;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	
	    $msg = "VM $vmid still running - terminating now with SIGTERM\n";
	    syslog ('info', $msg);
	    print $msg;
	    kill 15, $vzlist->{$vmid}->{pid};
	}

	# this is called by system shotdown scripts, so remaining
	# processes gets killed anyways (no need to send kill -9 here)

	$msg = "Qemu Server stopped\n";
	syslog ('info', $msg);
	print STDERR $msg;

    }
}

1;
